import sys, os, subprocess, fcntl, time, signal, shlex
from config import *

pipe_Learner = None
lst_GameProcess = []


#									
def run_process (_sCommand):
	args = shlex.split (_sCommand)
	pipe = subprocess.Popen (args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

	flags = fcntl.fcntl (pipe.stdout, fcntl.F_GETFL)
	fcntl.fcntl (pipe.stdout, fcntl.F_SETFL, flags | os.O_NONBLOCK)
	
	return pipe


#									
def run_support_process_in_shell (_sCommand):
	pipe = subprocess.Popen (_sCommand, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	lstLines = map (lambda x: x.rstrip (), pipe.stdout.readlines ())
	pipe.wait ()
	if 0 != len (lstLines):
		print '\n'.join (lstLines)


#									
def run_support_process (_sCommand):
	args = shlex.split (_sCommand)
	pipe = subprocess.Popen (args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	lstLines = map (lambda x: x.rstrip (), pipe.stdout.readlines ())
	pipe.wait ()
	if 0 != len (lstLines):
		print '\n'.join (lstLines)


#									
def run_game_process (_sName, _lstOptions):
	sLogId = 'none'
	if type(tuple()) == type(_lstOptions):
		sLogId = _lstOptions[0].rsplit('/',1)[-1]
	elif type('') == type(_lstOptions):
		sLogId = _lstOptions.rsplit('/',1)[-1]

	sLogPath = get_config (_sName + ":log_path")
	sLogFile = sLogPath + sLogId + '.log'
	sCommand = get_config (_sName + ":cmd")%_lstOptions + ' 2>&1 > ' + sLogFile
	return run_process (sCommand)


#									
def run_game ():
	global lst_GameProcess

	iServerPortBase = int (get_config ("server_port_base"))
	iClientPortBase = int (get_config ("client_port_base"))
	iParallelization = int (get_config ("simulator_count"))

	lst_GameProcess.append ((run_game_process ('server', (str(iServerPortBase))), 'game server'))
	lst_GameProcess.append ((run_game_process ('client', (str(iServerPortBase), \
											  '/tmp/' + str(iClientPortBase))), 'game client'))

	for i in xrange (1, iParallelization + 1):
		time.sleep (1)
		sServerPort = '/tmp/' + str(iServerPortBase + i)
		sClientPort = '/tmp/' + str(iClientPortBase + i)
		lst_GameProcess.append ((run_game_process ('server', (sServerPort)), ('server', sServerPort, 'simulation server')))
		lst_GameProcess.append ((run_game_process ('client', (sServerPort, sClientPort)),
											 ('client', (sServerPort, sClientPort), 'simulation client')))


#									
def run_learner (_sCommand):
	global pipe_Learner
	pipe_Learner = run_process (_sCommand)


#									
def monitor_game ():
	global lst_GameProcess
	bGameHasQuit = False
	for g, (pipe, data) in enumerate (lst_GameProcess):
		bHasQuit = (None != pipe.poll ())
		'''
		if True == bHasQuit:
			print '   game failure : ', data
			if 3 == len(data):
				print '   game failure attemting to restart : ', data
				lst_GameProcess [g] = (run_game_process (data[0], (data[1])), data)
				bHasQuit = False
			else:
				print '   game failure : ', data
		'''
		bGameHasQuit |= bHasQuit
	return bGameHasQuit


#									
def monitor_learner ():
	global pipe_Learner
	return (None != pipe_Learner.poll ())


#									
def stop_game ():
	global lst_GameProcess
	iProcessesEnded = 0
	while iProcessesEnded < len (lst_GameProcess):
		iProcessesEnded = 0
		for pipe, name in lst_GameProcess:
			if None != pipe.poll ():
				iProcessesEnded += 1
				continue
			try:
				os.kill (pipe.pid, signal.SIGTERM)
			except:
				None
		time.sleep (2)
	
	lst_GameProcess = []


#									
def stop_learner ():
	global pipe_Learner
	while None == pipe_Learner.poll ():
		try:
			os.kill (pipe_Learner.pid, signal.SIGTERM)
		except:
			None
		time.sleep (2)
	pipe_Learner = None


#									
def main ():
	load_config (sys.argv)

	file = open (get_config ('run_spec_file'))
	lstRuns = map (lambda x: x.strip (), file.readlines ())
	file.close ()

	#file = open (get_config ('run_log_file'), 'w')
	lstCommonSupportProcs = []
	lstSupportProcs = []
	sCommonConfig = ''
	for sRun in lstRuns:
		if '' == sRun:
			continue
		if '#' == sRun[0]:
			continue

		if True == sRun.startswith ('!*'):
			sCommonConfig = sRun [2:].strip ()
			#print sCommonConfig
			continue
		if True == sRun.startswith ('!'):
			lstCommonSupportProcs.append (sRun)
			continue

		if '*' != sRun[0]:
			lstSupportProcs.append (sRun)
			continue

		sRun = sRun[1:].strip ()
		print '> ' + sRun

		(sPrefix, sConfig, sLog) = sRun.split ("'")

		sOutputDir = os.path.dirname (sLog.strip ())
		if False == os.path.exists (sOutputDir):
			os.makedirs (sOutputDir)

		sRunDir = sLog.split ('/')[1]
		sParamsDir = sRunDir

		sInstanceConfig = sCommonConfig
		sInstanceConfig = sInstanceConfig.replace ('__outdir__', sOutputDir)
		sInstanceConfig = sInstanceConfig.replace ('__paramdir__', sParamsDir)
		sConfig += ' ' + sInstanceConfig + ' run=' + sRunDir
		sRun = sPrefix + "'" + sConfig + "'" + sLog

		for sSupport in lstCommonSupportProcs:
			if True == sSupport.startswith ('!s '):
				sSupport = sSupport[2:]
				sSupport = sSupport.strip ().replace ('__outdir__', sOutputDir)
				sSupport = sSupport.strip ().replace ('__paramdir__', sParamsDir)
				run_support_process_in_shell (sSupport)
			else:
				sSupport = sSupport[1:]
				sSupport = sSupport.strip ().replace ('__outdir__', sOutputDir)
				sSupport = sSupport.strip ().replace ('__paramdir__', sParamsDir)
				run_support_process (sSupport)

		for sSupport in lstSupportProcs:
			sSupport = sSupport.strip ().replace ('__outdir__', sOutputDir)
			sSupport = sSupport.strip ().replace ('__paramdir__', sParamsDir)
			run_support_process (sSupport)
		lstSupportProcs = []

		run_game ()
		time.sleep (5)
		run_learner (sRun)

		bGameQuit = monitor_game ()
		bLearnerQuit = monitor_learner ()
		while False == bGameQuit and False == bLearnerQuit:
			time.sleep (2)
			bGameQuit = monitor_game ()
			if True == bGameQuit:
				time.sleep (5)
				bGameQuit = monitor_game ()
			bLearnerQuit = monitor_learner ()
			if True == bLearnerQuit:
				time.sleep (5)
				bLearnerQuit = monitor_learner ()

		if True == bLearnerQuit:
			print 'Experiment completed'
			#file.write ('Experiment completed : ' + sRun + '\n')
		elif True == bGameQuit:
			print 'Terminating experiment on game failure'
			#file.write ('Terminating experiment on game failure : ' + sRun + '\n')

		stop_learner ()
		stop_game ()
		time.sleep (5)

	#file.close ()



if __name__ == '__main__':
	sys.exit (main ())
